import asyncio
import datetime
import gzip
import json
import logging
import multiprocessing
import os
import platform
import shutil
import sys
import warnings
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Any

import httpx
import pandas as pd
import tqdm
from tenacity import (
    before_sleep,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_fixed,
    wait_exponential,
)
from tqdm.contrib.concurrent import process_map
from tqdm.contrib.logging import logging_redirect_tqdm

from gpas import __version__, misc
from gpas.misc import (
    DEFAULT_ENVIRONMENT,
    ENVIRONMENTS,
    ENVIRONMENTS_URLS,
    FILE_TYPES,
    GOOD_STATUSES,
)
from gpas.validation import build_validation_message, validate

logger = logging.getLogger(__name__)


def parse_token(token: Path) -> dict:
    return json.loads(Path(token).read_text())


def parse_mapping_csv(mapping_csv: Path) -> dict:
    df = pd.read_csv(mapping_csv)
    expected_columns = {
        "local_batch",
        "local_run_number",
        "local_sample_name",
        "gpas_batch",
        "gpas_run_number",
        "gpas_sample_name",
    }
    if not expected_columns.issubset(set(df.columns)):
        raise RuntimeError(f"One or more expected columns missing from mapping CSV")
    return df.set_index("gpas_sample_name")["local_sample_name"].to_dict()


def fetch_user_details(access_token, environment: ENVIRONMENTS) -> dict:
    """Test API authentication and fetch response from userOrgDtls endpoint"""
    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['ORDS']}/userOrgDtls"
    try:
        logging.debug(f"Fetching user details {endpoint=}")
        r = httpx.get(
            endpoint, headers={"Authorization": f"Bearer {access_token}"}, timeout=10
        )
        if not r.is_success:
            r.raise_for_status()
        result = r.json()
        logging.debug(f"{result=}")
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 400 and "API access" in e.response.json().get(
            "message"
        ):
            raise misc.AuthenticationError(
                f"Authentication failed. User lacks API permissions"
            ) from None
        elif e.response.status_code == 401:
            raise misc.AuthenticationError(
                f"Authentication failed. Check access token and environment"
            ) from None
        else:
            raise e from None
    # Query PyPI instead of GitHub? GitHub has hopelessly low unauthenticated rate limits
    # try:
    #     r = httpx.get(
    #         "https://api.github.com/repos/GlobalPathogenAnalysisService/gpas-cli/releases/latest"
    #     )
    #     tag = r.json().get("tag_name")
    #     if tag and tag != __version__:
    #         logging.warning(
    #             f"Installed gpas-cli version ({__version__}) differs from the latest release ({tag})"
    #         )
    # except:
    #     pass
    return result


def parse_user_details(result: dict) -> tuple:
    """Parse response from userOrgDtls endpoint"""
    result = result.get("userOrgDtl", {})[0]
    user = result.get("userName")
    organisation = result.get("organisation")
    permitted_tags = [
        d.get("tagName") for d in result.get("tags", {}) if d.get("tagAccessYN") == "Y"
    ]
    logging.debug(f"{permitted_tags=}")
    date_mask = result.get("maskCollectionDate")  # Expects {"N", "MONTH", "WEEK"}
    return user, organisation, permitted_tags, date_mask


def update_fasta_header(path: Path, guid: str, name: str):
    """Update the header line of a gzipped fasta file in place"""
    with gzip.open(path, "rt") as fh:
        contents = fh.read()
    if guid in contents:
        with gzip.open(path, "wt") as fh:
            fh.write(contents.replace(guid, f"{guid}|{name}"))
    else:
        logging.warning(f"Could not rename {guid} inside {name}.fasta.gz")


async def fetch_status_async(
    access_token: str,
    guids: list | dict,
    environment: ENVIRONMENTS = DEFAULT_ENVIRONMENT,
) -> list[dict]:
    """Returns a list of dicts of containing status records"""
    fetch_user_details(access_token, environment)
    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }
    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_sample_detail"
    limits = httpx.Limits(
        max_keepalive_connections=10, max_connections=20, keepalive_expiry=10
    )
    transport = httpx.AsyncHTTPTransport(limits=limits, retries=0)
    async with httpx.AsyncClient(transport=transport, timeout=30) as client:
        guids_urls = {guid: f"{endpoint}/{guid}" for guid in guids}
        tasks = [
            fetch_status_single_async(client, guid, url, headers)
            for guid, url in guids_urls.items()
        ]
        records = [
            await f
            # for f in asyncio.as_completed(tasks)
            for f in tqdm.tqdm(
                asyncio.as_completed(tasks),
                desc=f"Querying status for {len(guids)} sample(s)",
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                total=len(tasks),
            )
        ]

    if type(guids) is dict:
        logging.debug("Renaming")
        records = pd.DataFrame(records).replace(guids).to_dict("records")

    return records


@retry(
    retry=retry_if_exception_type(httpx.HTTPError),
    wait=wait_exponential(multiplier=1, min=1, max=16),
    stop=stop_after_attempt(5),
    before_sleep=before_sleep.before_sleep_log(logger, 10),
)
async def fetch_status_single_async(client, guid, url, headers):
    logging.debug(f"fetch_status_single_async(): {url=}")
    r = await client.get(url=url, headers=headers)
    logging.debug(f"fetch_status_single_async(): {r.status_code=} {r.json()=}")
    if r.status_code == httpx.codes.OK:
        r_json = r.json()[0]
        status = r_json.get("status")
        result = dict(sample=guid, status=status)
        if status not in GOOD_STATUSES:
            with logging_redirect_tqdm():
                logging.info(f"{guid} has status {status}")
    elif r.status_code == 400 and "API access" in r.json().get("message"):
        raise misc.AuthenticationError(
            f"Authentication failed (HTTP {r.status_code}). User lacks API permissions"
        )
    elif r.status_code == 401:
        raise misc.AuthenticationError(
            f"Authentication failed (HTTP {r.status_code}). Invalid token?"
        )
    elif r.json().get("message") == "Sample not found.":
        status = "UNKNOWN"
        with logging_redirect_tqdm():
            logging.info(f"{guid} has status {status}")
        result = dict(sample=guid, status=status)
    elif r.json().get("message") == "You do not have access to this sample.":
        status = "UNAUTHORISED"
        with logging_redirect_tqdm():
            logging.info(f"{guid} has status {status}")
        result = dict(sample=guid, status=status)
    else:
        r.raise_for_status()  # Raises retryable httpx.HTTPError

    return result


async def download_async(
    access_token: str,
    guids: list | dict,
    file_types: list[str] = ["fasta"],
    out_dir: Path = Path.cwd(),
    environment: ENVIRONMENTS = DEFAULT_ENVIRONMENT,
):
    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }
    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_output"

    unrecognised_file_types = set(file_types) - {t.name for t in FILE_TYPES}
    if unrecognised_file_types:
        raise RuntimeError(f"Invalid file type(s): {unrecognised_file_types}")
    with logging_redirect_tqdm():
        logging.info(f"Fetching file types {file_types}")

    limits = httpx.Limits(
        max_keepalive_connections=5, max_connections=10, keepalive_expiry=10
    )
    transport = httpx.AsyncHTTPTransport(limits=limits, retries=0)
    async with httpx.AsyncClient(transport=transport, timeout=120) as client:
        guids_types_urls = {}
        for guid in guids:
            for file_type in file_types:
                guids_types_urls[(guid, file_type)] = f"{endpoint}/{guid}/{file_type}"
        tasks = [
            download_single_async(
                client,
                guid,
                file_type,
                url,
                headers,
                out_dir,
                guids[guid] if type(guids) is dict else None,
            )
            for (guid, file_type), url in guids_types_urls.items()
        ]
        return [
            await f
            for f in tqdm.tqdm(
                asyncio.as_completed(tasks),
                desc=f"Downloading {len(tasks)} files for {len(guids)} sample(s)",
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                total=len(tasks),
            )
        ]


@retry(
    retry=retry_if_exception_type(httpx.HTTPError),
    wait=wait_exponential(multiplier=1, min=1, max=8),
    stop=stop_after_attempt(4),
    before_sleep=before_sleep.before_sleep_log(logger, 10),
)
async def download_single_async(
    client, guid, file_type, url, headers, out_dir, name=None
):
    file_types_extensions = {
        "json": "json",
        "fasta": "fasta.gz",
        "bam": "bam",
        "vcf": "vcf",
    }
    prefix = name if name else guid
    r = await client.get(url=url, headers=headers)
    Path(out_dir).mkdir(parents=False, exist_ok=True)
    if r.status_code == httpx.codes.OK:
        logging.debug(
            Path(out_dir) / Path(f"{prefix}.{file_types_extensions[file_type]}")
        )
        with open(
            Path(out_dir) / Path(f"{prefix}.{file_types_extensions[file_type]}"), "wb"
        ) as fh:
            fh.write(r.content)
        if name and file_type == "fasta":
            update_fasta_header(
                Path(out_dir) / Path(f"{prefix}.{file_types_extensions[file_type]}"),
                guid,
                name,
            )
    elif r.status_code == 400 and "API access" in r.json().get("message"):
        raise misc.AuthenticationError(
            f"Bad request (HTTP {r.status_code}). User lacks API permissions"
        )
    elif r.status_code == 401:
        raise misc.AuthenticationError(
            f"Authorisation failed (HTTP {r.status_code}). Invalid token?"
        )
    else:
        r.raise_for_status()  # Raises retryable httpx.HTTPError


def fetch_status(
    access_token: str,
    guids: list | dict,
    environment: ENVIRONMENTS = DEFAULT_ENVIRONMENT,
) -> list[dict]:
    """
    Return a list of dictionaries given a list of guids
    """
    headers = {
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }
    endpoint = f"{ENVIRONMENTS_URLS[environment.value]['API']}/get_sample_detail"
    records = []
    for guid in tqdm.tqdm(guids):
        r = httpx.get(url=f"{endpoint}/{guid}", headers=headers)
        if r.is_success:
            if type(guids) == "dict":
                sample = guids[sample]
            else:  # list
                sample = r.json()[0].get("name")
            records.append({"sample": sample, "status": r.json()[0].get("status")})
        else:
            if type(guids) == "dict":
                sample = guids[sample]
            else:  # list
                sample = r.json()[0].get("name")
            records.append({"sample": sample, "status": "Unknown"})
            with logging_redirect_tqdm():
                logging.warning(f"{guid} (error {r.status_code})")
        if type(guids) is dict:
            records = pd.DataFrame(records).replace(guids).to_dict("records")

    return records


class Sample:
    """
    Represent a single sample
    """

    def __init__(
        self,
        batch,
        run_number,
        sample_name,
        control,
        collection_date,
        tags,
        country,
        region,
        specimen_organism,
        host,
        instrument_platform,
        primer_scheme,
        schema_name,
        working_dir,
        samtools_path,
        decontaminator_path,
        district=None,
        fastq=None,
        fastq1=None,
        fastq2=None,
        bam=None,
    ):
        self.batch = batch
        self.run_number = run_number
        self.gpas_run_number = None
        self.sample_name = sample_name
        self.fastq = fastq
        self.fastq1 = fastq1
        self.fastq2 = fastq2
        self.bam = bam
        self.control = control
        self.collection_date = collection_date
        self.tags = tags.split(":")
        self.country = country
        self.region = region
        self.district = district
        self.specimen_organism = specimen_organism
        self.host = host
        self.instrument_platform = instrument_platform
        self.primer_scheme = primer_scheme
        self.schema_name = schema_name
        self.paired = True if self.schema_name.startswith("Paired") else False
        self.decontamination_ref_path = self.get_decontamination_ref_path()
        self.working_dir = Path(working_dir)
        self.working_dir.mkdir(parents=False, exist_ok=True)
        logging.debug(f"{working_dir=}")
        self.guid = None
        self.mapping_path = None
        self.samtools_path = samtools_path
        self.decontaminator_path = decontaminator_path
        self.decontamination_stats = None

    def get_decontamination_ref_path(self):
        organisms_decontamination_references = {"SARS-CoV-2": "MN908947_no_polyA.fasta"}
        ref = organisms_decontamination_references[self.specimen_organism]
        return misc.get_data_path() / Path("refs") / Path(ref)

    def _get_decontaminate_cmd(self):
        if self.specimen_organism == "SARS-CoV-2":
            cmd = self._get_riak_cmd()
        else:
            raise misc.DecontaminationError("Invalid organism")

        command = misc.LoggedShellCommand(
            name=self.sample_name,
            action="decontamination",
            cmd=cmd,
        )
        return command

    def _get_convert_bam_cmd(self, paired=False) -> misc.LoggedShellCommand:
        prefix = Path(self.working_dir) / Path(self.sample_name)
        if not self.paired:
            cmd = f'"{self.samtools_path}" fastq -0 "{prefix}.fastq.gz" "{self.bam}"'
            self.fastq = self.working_dir / Path(self.sample_name + ".fastq.gz")
        else:
            cmd = (
                f'"{self.samtools_path}" sort -n "{self.bam}" |'
                f' "{self.samtools_path}" fastq -N'
                f' -1 "{prefix.parent / (prefix.name + "_1.fastq.gz")}"'
                f' -2 "{prefix.parent / (prefix.name + "_2.fastq.gz")}"'
                f' -s "{prefix.parent / (prefix.name + "_s.fastq.gz")}"'
            )
            self.fastq1 = self.working_dir / Path(self.sample_name + "_1.fastq.gz")
            self.fastq2 = self.working_dir / Path(self.sample_name + "_2.fastq.gz")

        command = misc.LoggedShellCommand(
            name=self.sample_name,
            action="bam_conversion",
            cmd=cmd,
        )

        return command

    def _get_riak_cmd(self) -> str:
        if not self.fastq2:
            cmd = (
                f'"{self.decontaminator_path}" --tech ont --enumerate_names'
                f' --ref_fasta "{self.decontamination_ref_path}"'
                f' --reads1 "{self.fastq}"'
                f' --outprefix "{self.working_dir / self.sample_name}"'
            )
        else:
            cmd = (
                f'"{self.decontaminator_path}" --tech illumina --enumerate_names'
                f' --ref_fasta "{self.decontamination_ref_path}"'
                f' --reads1 "{self.fastq1}"'
                f' --reads2 "{self.fastq2}"'
                f' --outprefix "{self.working_dir / self.sample_name}"'
            )
        self.clean_fastq = (
            self.working_dir / Path(self.sample_name + ".reads.fastq.gz")
            if self.fastq
            else None
        )
        self.clean_fastq1 = (
            self.working_dir / Path(self.sample_name + ".reads_1.fastq.gz")
            if self.fastq1
            else None
        )
        self.clean_fastq2 = (
            self.working_dir / Path(self.sample_name + ".reads_2.fastq.gz")
            if self.fastq2
            else None
        )
        return cmd

    def _hash_fastq(self):
        self.md5 = misc.hash_file(str(self.fastq))

    def _hash_fastqs(self):
        self.md5_1 = misc.hash_file(str(self.fastq1))
        self.md5_2 = misc.hash_file(str(self.fastq2))

    def _build_mapping_record(self) -> dict[str, Any]:
        return {
            "local_batch": self.batch,
            "local_run_number": self.run_number,
            "local_sample_name": self.sample_name,
            "gpas_run_number": self.gpas_run_number,
            "gpas_sample_name": self.guid,
        }


class Batch:
    """
    Represent a batch of samples
    """

    def __init__(
        self,
        upload_csv: Path,
        token: Path | None = None,
        working_dir: Path = Path("/tmp"),
        out_dir: Path = Path(),
        processes: int = 0,
        connections: int = 10,
        environment: ENVIRONMENTS = DEFAULT_ENVIRONMENT,
        json_messages: bool = False,
        save_reads: bool = False,
        user_agent_name: str = "",
        user_agent_version: str = "",
    ):
        self.upload_csv = upload_csv
        self.token = parse_token(token) if token else None
        self.environment = environment
        self.working_dir = working_dir
        self.out_dir = out_dir
        out_dir.mkdir(parents=False, exist_ok=True)
        self.processes = processes if processes else int(multiprocessing.cpu_count())
        self.connections = connections
        self.json = {"validation": "", "decontamination": "", "submission": ""}
        self.json_messages = json_messages
        self.samtools_path = misc.get_binary_path("samtools")
        self.decontaminator_path = misc.get_binary_path("readItAndKeep")
        self.save_reads = save_reads
        self.user_agent_name = user_agent_name
        self.user_agent_version = user_agent_version
        self.client_info = {
            "name": "gpas-cli",
            "version": __version__,
            "platform": platform.system(),
            "python_version": platform.python_version(),
        }
        if self.user_agent_name or self.user_agent_version:
            self.client_info["user_agent"] = {
                "name": self.user_agent_name,
                "version": self.user_agent_version,
            }
        if self.token:
            auth_result = fetch_user_details(
                self.token["access_token"], self.environment
            )
            (
                self.user,
                self.organisation,
                self.permitted_tags,
                self.date_mask,
            ) = parse_user_details(auth_result)
            self.headers = {
                "Authorization": f"Bearer {self.token['access_token']}",
                "Content-Type": "application/json",
                "charset": "utf-8",
            }

        else:
            self.user = None
            self.organisation = None
            self.permitted_tags = []
            self.headers = {}
            self.date_mask = None
        logging.debug(f"{self.upload_csv=}")
        self.df, self.schema = validate(self.upload_csv, self.permitted_tags)
        self.schema_name = self.schema.__schema__.name
        self.schema_fields = set(self.schema.to_schema().columns.keys()) | {
            "sample_name"
        }
        self.validation_json = build_validation_message(self.df, self.schema_name)
        batch_attrs = {
            "schema_name": self.schema_name,
            "working_dir": self.working_dir,
            "samtools_path": self.samtools_path,
            "decontaminator_path": self.decontaminator_path,
        }
        self.samples = [  # Pass only schematised fields to the Sample constructor
            Sample(
                **dict((k, r[k]) for k in self.schema_fields if k in r), **batch_attrs
            )
            for r in self.df.fillna("").reset_index().to_dict("records")
        ]
        self.paired = self.samples[0].paired

        self.uploaded_on = misc.oracle_timestamp()

        self._number_runs()
        self.uploaded = False
        if self.json_messages:
            misc.print_json(self.validation_json)

    def _get_convert_bam_cmds(self) -> list[misc.LoggedShellCommand]:
        return [s._get_convert_bam_cmd() for s in self.samples]

    def _get_decontaminate_cmds(self) -> list[misc.LoggedShellCommand]:
        return [s._get_decontaminate_cmd() for s in self.samples]

    def _save_reads(self) -> None:
        save_dir = self.out_dir / "decontaminated-reads"
        save_dir.mkdir(parents=False, exist_ok=True)
        for s in self.samples:
            if s.clean_fastq:
                shutil.copy2(s.clean_fastq, save_dir)
            if s.clean_fastq1:
                shutil.copy2(s.clean_fastq1, save_dir)
            if s.clean_fastq2:
                shutil.copy2(s.clean_fastq2, save_dir)
        logging.info(f"Saved decontaminated reads to {save_dir.resolve()}")

    def _decontaminate(self) -> None:
        """
        Convert BAM files to FASTQ files, if necessary, and then decontaminate
        """
        if "Bam" in self.schema_name:  # Conversion necessary
            misc.run_parallel_logged(
                self._get_convert_bam_cmds(),
                participle="Converting",
                json_messages=self.json_messages,
                processes=self.processes,
            )
        samples_runs = misc.run_parallel_logged(
            self._get_decontaminate_cmds(),
            participle="Decontaminating",
            json_messages=self.json_messages,
            processes=self.processes,
        )
        self._parse_decontamination_stats(samples_runs)
        if self.save_reads:
            self._save_reads()

    def _parse_decontamination_stats(self, samples_runs) -> None:
        samples_decontamination_stats = {
            s: parse_decontamination_stats(r.stdout) for s, r in samples_runs.items()
        }
        for s in self.samples:
            s.decontamination_stats = samples_decontamination_stats.get(s.sample_name)

    def _hash_fastqs(self):
        if not self.paired:
            list(map(lambda s: s._hash_fastq(), self.samples))
        else:
            list(map(lambda s: s._hash_fastqs(), self.samples))

    def _get_sample_attrs(self, attr) -> dict[str, Any]:
        return {s.sample_name: getattr(s, attr) for s in self.samples}

    def _set_samples(self, name, value):
        for s in self.samples:
            setattr(s, name, value)

    def _fetch_guids(self):
        md5_attr = "md5" if not self.paired else "md5_1"
        checksums = list(self._get_sample_attrs(md5_attr).values())
        payload = {
            "batch": {
                "organisation": self.organisation,
                "uploadedOn": self.uploaded_on,
                "uploadedBy": self.user,
                "samples": checksums,
            }
        }
        logging.debug(f"_fetch_guids(): {payload=}")
        endpoint = (
            f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/createSampleGuids"
        )
        logging.debug(f"Fetching guids; {endpoint=}")
        r = httpx.post(
            url=endpoint, data=json.dumps(payload), headers=self.headers, timeout=120
        )
        if not r.is_success:
            r.raise_for_status()
        result = r.json()
        logging.debug(f"{result=}")
        self.batch_guid = result["batch"]["guid"]
        guids_hashes = {s["guid"]: s["hash"] for s in result["batch"]["samples"]}
        logging.debug(f"{guids_hashes=}")

        # Used hash-keyed dict of lists to tolerate samples with same hash
        hashes_guids = defaultdict(list)
        for g, h in guids_hashes.items():
            hashes_guids[h].append(g)

        # Assign each sample a guid from the pool associated with each hash
        for sample in self.samples:
            sample.guid = hashes_guids[getattr(sample, md5_attr)].pop()

    def _rename_fastqs(self):
        """Rename decontaminated fastqs using server-side guids"""
        for s in self.samples:
            # Workaround for ReadItAndKeep making fastas when encountering empty fastqs
            if s.clean_fastq and not s.clean_fastq.exists():
                s.clean_fastq = s.working_dir / Path(s.sample_name + ".reads.fasta.gz")
            if s.clean_fastq1 and not s.clean_fastq1.exists():
                s.clean_fastq1 = s.working_dir / Path(
                    s.sample_name + ".reads_1.fasta.gz"
                )
            if s.clean_fastq2 and not s.clean_fastq2.exists():
                s.clean_fastq2 = s.working_dir / Path(
                    s.sample_name + ".reads_2.fasta.gz"
                )

            s.clean_fastq = (
                s.clean_fastq.rename(s.working_dir / Path(s.guid + ".reads.fastq.gz"))
                if s.clean_fastq
                else None
            )
            s.clean_fastq1 = (
                s.clean_fastq1.rename(
                    s.working_dir / Path(s.guid + ".reads_1.fastq.gz")
                )
                if s.clean_fastq1
                else None
            )
            s.clean_fastq2 = (
                s.clean_fastq2.rename(
                    s.working_dir / Path(s.guid + ".reads_2.fastq.gz")
                )
                if s.clean_fastq2
                else None
            )

    def _build_mapping_csv(self):
        records = [  # Collects attrs from Sample, except for gpas_batch (Batch)
            {**s._build_mapping_record(), "gpas_batch": self.batch_guid}
            for s in self.samples
        ]
        df = pd.DataFrame(
            records,
            columns=[
                "local_batch",
                "local_run_number",
                "local_sample_name",
                "gpas_batch",
                "gpas_run_number",
                "gpas_sample_name",
            ],
        )
        arbitrary_fields = [c for c in self.df.columns if c not in self.schema_fields]
        combined_df = pd.merge(
            df,
            self.df[[*arbitrary_fields]],
            left_on="local_sample_name",
            right_index=True,
        )
        target_path = self.out_dir / Path(self.batch_guid + ".mapping.csv")
        combined_df.to_csv(target_path, index=False)
        self.mapping_path = target_path
        logging.info(f"Saved mapping CSV to {self.mapping_path}")

    def _fetch_par(self):
        """Private method that calls ORDS to get a Pre-Authenticated Request.

        The PAR url is used to upload data to the Organisation's input bucket in OCI

        Returns
        -------
        par: str
        """
        endpoint = f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/pars"
        logging.debug(f"Fetching PAR; {endpoint=} {self.headers=}")
        r = httpx.get(url=endpoint, headers=self.headers)
        if not r.is_success:
            r.raise_for_status()
        result = json.loads(r.content)
        logging.debug(f"{result=}")
        if result.get("status") == "error":
            raise RuntimeError("Problem fetching PAR")
        logging.debug(f"{result}, {r.status_code}")
        self.par = result["par"]
        self.bucket = self.par.split("/")[-3]
        self.batch_url = self.par + self.batch_guid + "/"

    def _get_uploads(self) -> list[misc.SampleUpload]:
        """Return local file path and destination URL for the file to be uploaded"""
        uploads = []
        for s in self.samples:
            url_prefix = self.batch_url + s.guid
            if not s.paired:
                uploads.append(
                    misc.SampleUpload(
                        name=s.sample_name,
                        path1=s.clean_fastq,
                        path2=None,
                        url1=f"{url_prefix}.reads.fastq.gz",
                        url2=None,
                    )
                )
            else:
                uploads.append(
                    misc.SampleUpload(
                        name=s.sample_name,
                        path1=s.clean_fastq1,
                        path2=s.clean_fastq2,
                        url1=f"{url_prefix}.reads_1.fastq.gz",
                        url2=f"{url_prefix}.reads_2.fastq.gz",
                    )
                )
        return uploads

    def _upload_samples(self) -> None:
        if self.json_messages:
            misc.print_progress_message_json(action="upload", status="started")
        uploads = self._get_uploads()
        if self.connections == 1:
            logging.debug("Single upload process")
            for upload in uploads:
                misc.upload_sample(
                    upload=upload,
                    headers=self.headers,
                    json_messages=self.json_messages,
                )
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")  # Suppress TqdmWarning about chunksize
                process_map(
                    partial(
                        misc.upload_sample,
                        headers=self.headers,
                        json_messages=self.json_messages,
                    ),
                    uploads,
                    max_workers=self.connections,
                    desc=f"Uploading {len(uploads)} sample(s)",
                    bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                    leave=False,
                )
        self.uploaded = True
        if self.json_messages:
            misc.print_progress_message_json(action="upload", status="finished")
        else:
            logging.info(f"Finished uploading {len(uploads)} sample(s)")

    def _finalise_submission(self):
        @retry(
            retry=retry_if_exception_type(httpx.HTTPError),
            wait=wait_fixed(30),
            stop=stop_after_attempt(2),
            before_sleep=before_sleep.before_sleep_log(logger, 10),
        )
        def post_submission(submission: dict, headers: dict):
            """Submit sample metadata to batches endpoint"""
            endpoint = f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/batches"
            logging.debug(f"post_submission(): {json.dumps(self.submission, indent=4)}")
            r = httpx.post(
                url=endpoint,
                data=json.dumps(submission, ensure_ascii=False).encode("utf-8"),
                headers=headers,
                timeout=180,
            )
            logging.debug(f"post_submission(): {r.text=}")
            r.raise_for_status()
            if r.json().get("status") != "success":
                raise misc.SubmissionError(r.json().get("errorMsg"))

        @retry(
            retry=retry_if_exception_type(httpx.HTTPError),
            wait=wait_fixed(10),
            stop=stop_after_attempt(2),
            before_sleep=before_sleep.before_sleep_log(logger, 10),
        )
        def put_done_mark(batch_guid: str, headers: dict):
            """Put upload done marker"""
            url = self.par + batch_guid + "/upload_done.txt"
            logging.debug(f"put_done_mark(): {url=}")
            r = httpx.put(url=url, headers=self.headers)
            logging.debug(f"put_done_mark(): {r.text=}")
            r.raise_for_status()

        if not self.uploaded:
            raise RuntimeError("Reads not uploaded")

        self.submission["batch"]["uploader"]["upload_finish"] = misc.oracle_timestamp()
        post_submission(self.submission, self.headers)
        put_done_mark(self.batch_guid, self.headers)
        logging.info(f"Submitted batch {self.batch_guid}")
        success_message = {
            "submission": {
                "status": "success",
                "batch": self.batch_guid,
                "samples": [s.sample_name for s in self.samples],
            }
        }
        if self.json_messages:
            misc.print_json(success_message)

    def _submit(self):
        self._fetch_par()
        self._upload_samples()
        self._prepare_submission()
        self._finalise_submission()

    def _prepare_submission(self):
        """Prepare the JSON payload for the GPAS Upload app

        Returns
        -------
            dict : JSON payload to pass to GPAS Electron upload app via STDOUT
        """

        samples = []
        for s in self.samples:
            if self.date_mask in {"WEEK", "MONTH"}:
                dt = datetime.datetime.strptime(s.collection_date, "%Y-%m-%d")
                if self.date_mask == "MONTH":
                    s.collection_date = dt.replace(day=1).strftime("%Y-%m-%d")
                elif self.date_mask == "WEEK":
                    td = datetime.timedelta(days=dt.weekday())
                    s.collection_date = (dt - td).strftime("%Y-%m-%d")
                logging.info(
                    f"Masked collection dates to start of {self.date_mask.lower()}"
                )
            sample = {
                "name": s.guid,
                "run_number": s.gpas_run_number,
                "tags": s.tags,
                "control": s.control,
                "collection_date": s.collection_date,
                "country": s.country,
                "region": s.region,
                "district": s.district,
                "specimen": s.specimen_organism,
                "host": s.host,
                "instrument": {"platform": s.instrument_platform},
                "primer_scheme": s.primer_scheme,
                "decontamination_stats": s.decontamination_stats,
            }
            if self.paired:
                sample["pe_reads"] = {
                    "r1_uri": str(s.clean_fastq1),
                    "r1_md5": s.md5_1,
                    "r2_uri": str(s.clean_fastq2),
                    "r2_md5": s.md5_2,
                }
                logging.debug(f"{s.clean_fastq1=}, {s.clean_fastq2=}")
            else:
                sample["se_reads"] = {"uri": str(s.clean_fastq), "md5": s.md5}
            samples.append(sample)

        self.submission = {
            "status": "completed",
            "batch": {
                "file_name": self.batch_guid,
                "bucket_name": self.bucket,
                "uploaded_on": self.uploaded_on,
                "uploaded_by": self.user,
                "organisation": self.organisation,
                "run_numbers": self.run_numbers,
                "samples": samples,
                "uploader": {
                    **self.client_info,
                    "upload_start": self.uploaded_on,
                    "upload_finish": None,  # Set in _finalise_submission()
                },
            },
        }

    def post_exception(self, exception: Exception):
        """Report exceptions occurring during Batch.upload()"""
        try:
            if "PYTEST_CURRENT_TEST" in os.environ:  # Disable reporting under pytest
                return
            e_t, e_v, e_tb = misc.get_value_traceback_fmt(exception)
            endpoint = (
                f"{ENVIRONMENTS_URLS[self.environment.value]['ORDS']}/logUploaderError"
            )
            payload = {
                "exception": {
                    "class": e_t,
                    "message": e_v,
                    "traceback": e_tb,
                    "timestamp": misc.oracle_timestamp(),
                },
                "uploader": self.client_info,
            }
            logging.debug(f"Exception payload {payload=}")
            r = httpx.post(url=endpoint, json=payload, headers=self.headers)
            logging.debug(f"Exception submission {r.is_success=}")
        except Exception:
            pass

    def upload(self, dry_run: bool = False, save_reads: bool = False) -> None:
        try:
            logging.info(
                f"Using {self.processes} process(es), {self.connections} connection(s)"
            )
            self._decontaminate()
            if not self.headers:
                logging.warning("No token provided, quitting")
                sys.exit()
            self._hash_fastqs()
            self._fetch_guids()
            self._build_mapping_csv()
            self._rename_fastqs()
            if dry_run:
                logging.info("Skipped submission")
            else:
                self._submit()
        except Exception as e:
            self.post_exception(e)
            raise e

    def _number_runs(self) -> None:
        """Enumerate unique values of run_number for submission"""
        samples_runs = self._get_sample_attrs("run_number")
        logging.debug(f"{samples_runs=}")
        samples_run_numbers = misc.number_runs(samples_runs)
        logging.debug(f"{samples_run_numbers=}")
        for s in self.samples:
            s.gpas_run_number = samples_run_numbers[s.sample_name]
        self.run_numbers = list(filter(None, samples_run_numbers.values()))
        logging.debug(f"{self.run_numbers=}")


def parse_decontamination_stats(stdout: str) -> dict:
    """
    Parse read-it-and-keep kept and discarded read counts
    """
    lines = stdout.strip().splitlines()
    counts = [int(l.rpartition("\t")[2]) for l in lines]
    count_in = counts[0] + counts[1]
    count_out = counts[2] + counts[3]
    delta = count_in - count_out
    assert delta >= 0
    try:
        fraction = round(delta / count_in, 4)
    except ArithmeticError:  # ZeroDivisionError
        fraction = 0
    return {
        "in": count_in,
        "out": count_out,
        "fraction": fraction,
    }



import json
import pycountry

# import pandas as pd

d = {
    c.alpha_3: [s.name for s in pycountry.subdivisions.get(country_code=c.alpha_2)]
    for c in pycountry.countries
}
with open("countries_subdivisions.json", "w", encoding="utf-8") as fh:
    json.dump(d, fh, ensure_ascii=False)

# pd.DataFrame([{'country': pycountry.countries.get(alpha_2=s.country_code).name, 'alpha_3': pycountry.countries.get(alpha_2=s.country_code).alpha_3, 'subdivision': s.name} for s in pycountry.subdivisions]).to_csv('countries_subdivisions.csv', index=False)



import subprocess
from pathlib import Path

import pandas as pd
import pytest
from gpas import lib
from gpas.misc import ENVIRONMENTS

data_dir = "tests/test-data"


def run(cmd, cwd="./"):  # Helper for CLI testing
    return subprocess.run(
        cmd, cwd=data_dir, shell=True, check=True, text=True, capture_output=True
    )


auth = lib.parse_token(Path(data_dir) / Path("token.json"))
auth_result = lib.fetch_user_details(auth["access_token"], ENVIRONMENTS.dev)
_, _, permitted_tags, _ = lib.parse_user_details(auth_result)


def test_upload_ont_bam_dry():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json large-nanopore-bam.csv --dry-run"
    )
    assert "INFO: Finished converting 1 sample(s)" in run_cmd.stderr
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr
    batch_guid = str(next(Path(data_dir).glob("*.mapping.csv"))).partition(".")[0]
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_ont_bam_dry_json_and_mapping_csv():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json large-nanopore-bam.csv --dry-run --json-messages"
    )
    assert "bam_conversion" in run_cmd.stdout
    assert "COVID_locost_2_barcode10" in run_cmd.stdout
    assert "finished" in run_cmd.stdout
    assert "decontamination" in run_cmd.stdout

    # Check all mapping CSV fields except for server side names
    batch_guid = str(next(Path(data_dir).glob("*.mapping.csv"))).partition(".")[0]
    records = pd.read_csv(Path(f"{batch_guid}.mapping.csv"), dtype=str).to_dict(
        "records"
    )
    assert records[0]["local_batch"] == "run1"
    assert records[0]["local_run_number"] == "run1.1"
    assert records[0]["local_sample_name"] == "COVID_locost_2_barcode10"
    assert records[0]["gpas_run_number"] == "1"
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_dry_working_dir():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json large-illumina-bam.csv --dry-run --working-dir temp"
    )
    assert (data_dir / Path("temp") / Path("cDNA-VOC-1-v4-1_2.fastq.gz")).exists()
    run(f"rm -rf temp *.mapping.csv")


def test_upload_action_level_messages():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json --dry-run --json-messages --processes 1 large-illumina-bam.csv"
    )
    assert (
        """{
    "progress": {
        "action": "bam_conversion",
        "status": "started"
    }
}
{
    "progress": {
        "action": "bam_conversion",
        "status": "started",
        "sample": "cDNA-VOC-1-v4-1"
    }
}
{
    "progress": {
        "action": "bam_conversion",
        "status": "finished",
        "sample": "cDNA-VOC-1-v4-1"
    }
}
{
    "progress": {
        "action": "bam_conversion",
        "status": "finished"
    }
}
{
    "progress": {
        "action": "decontamination",
        "status": "started"
    }
}
{
    "progress": {
        "action": "decontamination",
        "status": "started",
        "sample": "cDNA-VOC-1-v4-1"
    }
}
{
    "progress": {
        "action": "decontamination",
        "status": "finished",
        "sample": "cDNA-VOC-1-v4-1"
    }
}
{
    "progress": {
        "action": "decontamination",
        "status": "finished"
    }
}"""
        in run_cmd.stdout
    )
    run(f"rm -rf *.mapping.csv")


def test_upload_empty():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json empty-fastq.csv --dry-run"
    )
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr
    batch_guid = run_cmd.stderr.partition("saved to ")[2].partition(".mapping.csv")[0]
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_empty_paired():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json empty-paired-fastq.csv --dry-run"
    )
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr
    batch_guid = run_cmd.stderr.partition("saved to ")[2].partition(".mapping.csv")[0]
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_empty_after_decontamination():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json empty-after-decontamination-fastq.csv --dry-run --connections 1"
    )
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr
    batch_guid = run_cmd.stderr.partition("saved to ")[2].partition(".mapping.csv")[0]
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_empty_after_decontamination_paired():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json empty-after-decontamination-paired-fastq.csv --dry-run --connections 2"
    )
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr
    batch_guid = run_cmd.stderr.partition("saved to ")[2].partition(".mapping.csv")[0]
    run(f"rm -f {batch_guid}.mapping.csv")


def test_upload_dots_in_sample_name():
    run_cmd = run(f"gpas upload dots-in-sample-name.csv")



import json
import subprocess
from pathlib import Path

import pytest
from gpas import data_dir, lib, misc, validation

data_dir = "tests/test-data"


def run(cmd, cwd="./"):  # Helper for CLI testing
    return subprocess.run(
        cmd, cwd=data_dir, shell=True, check=True, text=True, capture_output=True
    )


def test_version():
    run_cmd = run("gpas --version")


def test_static_assets_exist():
    validation.parse_countries_subdivisions()
    assert misc.get_reference_path("SARS-CoV-2").exists()


def test_validate_ok():
    df, schema = validation.validate(Path(data_dir) / Path("large-illumina-fastq.csv"))
    schema_name = schema.__schema__.name
    message = validation.build_validation_message(df, schema_name)
    # assert message == {
    #     "validation": {
    #         "status": "success",
    #         # "schema": "PairedFastqSchema",
    #         "samples": [
    #             {
    #                 "sample_name": "cDNA-VOC-1-v4-1",
    #                 "files": [
    #                     "/Users/bede/Research/Git/gpas-cli/tests/test-data/reads/large-illumina-fastq_1.fastq.gz",
    #                     "/Users/bede/Research/Git/gpas-cli/tests/test-data/reads/large-illumina-fastq_2.fastq.gz",
    #                 ],
    #             }
    #         ],
    #     }
    # }
    # Can't do above because of absolute paths
    assert message["validation"]["status"] == "success"
    assert message["validation"]["samples"][0]["sample_name"] == "cDNA-VOC-1-v4-1"
    assert message["validation"]["samples"][0]["files"][0].endswith(
        "reads/large-illumina-fastq_1.fastq.gz"
    )
    assert message["validation"]["samples"][0]["files"][1].endswith(
        "reads/large-illumina-fastq_2.fastq.gz"
    )


def test_validate_new_cli():
    run_cmd = run("gpas validate broken/impossible-country-region.csv --json-messages")
    assert "cDNA-VOC-1-v4-1" in run_cmd.stdout
    assert (
        "invalid region (ISO 3166-2 subdivision) for specified country"
        in run_cmd.stdout
    )


def test_validate_fail_no_tags():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("no-tags.csv")
        )
    assert e.value.errors == [
        {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be empty"}
    ]
    assert e.value.report == {
        "validation": {
            "status": "failure",
            "errors": [
                {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be empty"}
            ],
        }
    }


def test_validate_fail_no_tags_colon():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("no-tags-colon.csv")
        )
    assert e.value.errors == [
        {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be empty"}
    ]


def test_validate_fail_dupe_tags():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("dupe-tags.csv")
        )
    assert e.value.errors == [
        {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be repeated"}
    ]


def test_validate_wrong_tags():
    """Should pass since tag validation does not happen offline"""
    run_cmd = run(f"gpas validate broken/wrong-tags.csv")
    assert "Validation successful" in run_cmd.stderr
    assert run_cmd.stdout == ""


def test_validate_fail_missing_files():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("broken-path.csv")
        )
    assert e.value.errors == [
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq1 file does not exist",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq2 file does not exist",
        },
    ]


def test_validate_fail_different_platforms():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("different-platforms.csv")
        )
    assert e.value.errors == [
        {
            "error": "instrument_platform must be the same for all samples in a submission"
        }
    ]


def test_validate_fail_country_region():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("invalid-country-region.csv")
        )
    assert e.value.errors == [
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "Bretagn is not a valid ISO 3166-2 subdivision name",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "US is not a valid ISO 3166-1 alpha-3 country code",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "invalid region (ISO 3166-2 subdivision) for specified country",
        },
    ]


def test_validate_empty_region():
    """Empty region should be fine"""
    _, message = validation.validate(Path(data_dir) / Path("empty-region.csv"))


def test_validate_empty_district():
    """Empty district should be fine"""
    _, message = validation.validate(Path(data_dir) / Path("empty-district.csv"))


def test_validate_empty_district_multiple():
    """Checking that a numerical district is correctly cast to a string by read_csv dtype"""
    _, message = validation.validate(
        Path(data_dir) / Path("empty-district-multiple.csv")
    )


def test_validate_fail_select_schema():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("no-schema.csv")
        )
    assert e.value.errors == [
        {
            "error": "could not infer upload CSV schema. For Nanopore samples, column 'instrument_platform' must be 'Nanopore', and either column 'fastq' or column 'bam' must be valid paths. For Illumina samples, column 'instrument_platform' must be 'Illumina' and either columns 'bam' or 'fastq1' and 'fastq2' must be valid paths."
        }
    ]


def test_validate_fail_wrong_instrument():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("wrong-instrument.csv")
        )
    assert e.value.errors == [
        {
            "error": "could not infer upload CSV schema. For Nanopore samples, column 'instrument_platform' must be 'Nanopore', and either column 'fastq' or column 'bam' must be valid paths. For Illumina samples, column 'instrument_platform' must be 'Illumina' and either columns 'bam' or 'fastq1' and 'fastq2' must be valid paths."
        }
    ]


def test_validate_illumina_fastq_wrong_platform():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir)
            / Path("broken")
            / Path("large-illumina-fastq-wrong-platform.csv")
        )
    assert e.value.errors == [
        {
            "error": "could not infer upload CSV schema. For Nanopore samples, column 'instrument_platform' must be 'Nanopore', and either column 'fastq' or column 'bam' must be valid paths. For Illumina samples, column 'instrument_platform' must be 'Illumina' and either columns 'bam' or 'fastq1' and 'fastq2' must be valid paths."
        }
    ]


def test_validate_fail_path_suffix():
    """Check that multiple errors are caught in one go (laziness)"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("bad-path-suffix.csv")
        )
    assert e.value.errors == [
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq1 file does not exist",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq2 file does not exist",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq2 must end with .fastq.gz or .bam as appropriate",
        },
    ]


def test_validate_nullable_batch():
    _, message = validation.validate(Path(data_dir) / Path("empty-batch-run.csv"))


def test_validate_nullable_run():
    _, message = validation.validate(Path(data_dir) / Path("empty-run.csv"))


def test_decontamination():
    lib.Batch(Path(data_dir) / Path("large-nanopore-fastq.csv"))._decontaminate()
    lib.Batch(Path(data_dir) / Path("large-illumina-fastq.csv"))._decontaminate()
    lib.Batch(Path(data_dir) / Path("large-nanopore-bam.csv"))._decontaminate()
    lib.Batch(Path(data_dir) / Path("large-illumina-bam.csv"))._decontaminate()


def test_decontamination_effectiveness():
    batch = lib.Batch(Path(data_dir) / Path("contaminated-nanopore-fastq.csv"))
    batch._decontaminate()
    for sample in batch.samples:
        assert sample.decontamination_stats == {"fraction": 0.5, "in": 2, "out": 1}


def test_decontamination_stats():
    stdout = """Input reads file 1	5034
Input reads file 2	5034
Kept reads 1	5006
Kept reads 2	5006

"""
    assert lib.parse_decontamination_stats(stdout) == {
        "in": 10068,
        "out": 10012,
        "fraction": 0.0056,
    }


def test_decontamination_stats_zero():
    stdout = """Input reads file 1	0
Input reads file 2	0
Kept reads 1	0
Kept reads 2	0

"""
    assert lib.parse_decontamination_stats(stdout) == {
        "in": 0,
        "out": 0,
        "fraction": 0,
    }


def test_upload_no_token_save_reads():
    """When run without a token, upload should quit after decontamination"""
    run_cmd = run("gpas upload large-nanopore-fastq.csv --save-reads")
    assert "Saved decontaminated reads" in run_cmd.stderr
    assert (
        data_dir
        / Path("decontaminated-reads")
        / Path("COVID_locost_2_barcode10.reads.fastq.gz")
    ).exists()
    run("rm -rf decontaminated-reads")


def test_weird_illumina_suffix():
    run("cp reads/large-illumina-fastq_1.fastq.gz reads/foo.fastq.gz")
    run("cp reads/large-illumina-fastq_2.fastq.gz reads/baa.fastq.gz")
    _, message = validation.validate(Path(data_dir) / Path("weird-illumina-suffix.csv"))
    run("rm reads/foo.fastq.gz reads/baa.fastq.gz")


def test_fail_no_header_cli():
    """Generic uncaught exception"""
    run_cmd = run("gpas validate broken/no-header.csv --json-messages")
    assert "unsupported operand type(s) for -: 'str' and 'int'" in run_cmd.stdout


def test_space_mapping_csv_reads():
    """Spaces should be tolerated in mapping CSV and reads filenames"""
    df, schema_name = validation.validate(
        Path(data_dir) / Path("large-nanopore-fastq space.csv")
    )


def test_leading_space_mapping_csv_reads():
    """Leading spaces should be tolerated in mapping CSV and reads filenames"""
    df, schema_name = validation.validate(
        Path(data_dir) / Path(" leading-space-large-nanopore-fastq.csv")
    )


def test_validate_fail_fastq_empty():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-fastq.csv")
        )
    assert e.value.errors[0]["error"] == "fastq cannot be empty"


def test_validate_fail_bam_empty():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-bam.csv")
        )
    assert e.value.errors[0]["error"] == "bam cannot be empty"


def test_validate_fail_completely_empty():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty.csv")
        )
    assert (
        e.value.errors[0]["error"]
        == "failed to parse upload CSV (No columns to parse from file)"
    )


def test_validate_fail_invalid_date():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("invalid-date.csv")
        )
    assert len(e.value.errors) == 1
    assert (
        "collection_date must be in format YYYY-MM-DD between 2019-01-01"
        in e.value.errors[0]["error"]
    )


def test_validate_fail_non_iso_date():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("non-iso-date.csv")
        )
    assert len(e.value.errors) == 1
    assert (
        "collection_date must be in format YYYY-MM-DD between 2019-01-01"
        in e.value.errors[0]["error"]
    )


def test_validate_fail_empty_date():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-date.csv")
        )

    assert len(e.value.errors) == 1
    assert e.value.errors == [
        {
            "sample_name": "COVID_locost_2_barcode10",
            "error": "collection_date cannot be empty",
        }
    ]


def test_validate_fail_insane_date():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("insane-date.csv")
        )
    assert len(e.value.errors) == 1
    assert (
        "collection_date must be in format YYYY-MM-DD between 2019-01-01"
        in e.value.errors[0]["error"]
    )


def test_validate_fail_dupe_fastqs():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("dupe-fastqs.csv")
        )
    assert len(e.value.errors) == 1
    assert e.value.errors[0] == {
        "sample_name": "COVID_locost_2_barcode10_x",
        "error": "fastq must be unique",
    }


def test_validate_fail_dupe_names():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("dupe-names.csv")
        )
    assert len(e.value.errors) == 1
    assert e.value.errors[0] == {
        "error": "sample_name must be unique",
    }


def test_validate_fail_empty_name():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-name.csv")
        )
    assert e.value.errors[0] == {"error": "sample_name cannot be empty"}


def test_validate_cayman():
    _, message = validation.validate(Path(data_dir) / Path("cayman.csv"))


def test_validate_vietnam():
    _, message = validation.validate(Path(data_dir) / Path("vietnam.csv"))


def test_validate_guyana():
    """Contains a space in the region field"""
    _, message = validation.validate(Path(data_dir) / Path("guyana.csv"))


def test_validate_no_run_batch():
    _, message = validation.validate(
        Path(data_dir) / Path("large-nanopore-fastq-no-run-batch.csv")
    )


def test_validate_negative():
    _, message = validation.validate(Path(data_dir) / Path("negative-control.csv"))


def test_validate_fail_not_unicode():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("not-unicode.csv")
        )
    assert "failed to parse upload CSV" in e.value.errors[0]["error"]


def test_paired_bam_first_read_not_equal():
    """Ensure that R1 is not the same as R2"""
    run_cmd = run("gpas upload large-illumina-bam.csv --working-dir ./")
    r1_str = run("zcat < cDNA-VOC-1-v4-1_1.fastq.gz | head -n 2 | tail -n 1").stdout
    r2_str = run("zcat < cDNA-VOC-1-v4-1_2.fastq.gz | head -n 2 | tail -n 1").stdout
    assert r1_str != r2_str
    run(
        "rm cDNA-VOC-1-v4-1_1.fastq.gz cDNA-VOC-1-v4-1_2.fastq.gz cDNA-VOC-1-v4-1.reads_1.fastq.gz cDNA-VOC-1-v4-1.reads_2.fastq.gz cDNA-VOC-1-v4-1_s.fastq.gz"
    )


def test_validate_fail_illegal_chars():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("illegal-chârs.csv")
        )
    assert e.value.errors[0] == {
        "error": "upload csv path contains illegal characters",
    }


def test_validate_fail_dupe_fastqs_illumina():
    """Two records, one with duplicate fastq1"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("dupe-fastqs-illumina.csv")
        )
    assert e.value.errors == [
        {
            "sample_name": "cDNA-VOC-1-v4-1",
            "error": "fastq1 and fastq2 must be jointly unique",
        },
        {
            "sample_name": "cDNA-VOC-1-v4-2",
            "error": "fastq1 and fastq2 must be jointly unique",
        },
        {"sample_name": "cDNA-VOC-1-v4-2", "error": "fastq1 must be unique"},
        {"sample_name": "cDNA-VOC-1-v4-2", "error": "fastq2 must be unique"},
    ]


def test_validate_fail_dupe_fastq1_fastq2_illumina():
    """Two records, one with duplicate fastq1"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("dupe-fastq1-fastq2.csv")
        )
    assert e.value.errors[0] == {
        "sample_name": "cDNA-VOC-1-v4-1",
        "error": "fastq1 and fastq2 cannot be the same",
    }


def test_validate_upload_path_windows():
    """Test that Windows paths are valid"""
    assert validation.validate_upload_csv_path(Path("C:\\test\\file\\path.jpg"))


def test_upload_no_token_user_agent():
    """Check that user agent name and version can be specified"""
    run_cmd = run(
        "gpas upload large-nanopore-fastq.csv --user-agent-name ClientyMcClientFace --user-agent-version 0.1.2"
    )


def test_validate_fail_epochalypse():
    """Test future date after UNIX epochalypse"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("epochalypse.csv")
        )


def test_upload_improperly_paired():
    """Check that improperly paired bams including properly paired reads do not trigger a crash"""
    run_cmd = run("gpas upload broken/improperly-paired-bam.csv")


def test_auth_response_parsing():
    json_payload = """{
	"userOrgDtl": [{
		"userName": "BEDE.CONSTANTINIDES@NDM.OX.AC.UK",
		"orgGUID": 261299400571842546812760352566796398698,
		"organisation": "University of Oxford",
		"maskCollectionDate": "N",
		"autoReleasedYN": "N",
		"autoApproveYN": "N",
		"tags": [{
			"tagName": "Positive_control",
			"tagDescription": "Fictitious tag for positive control samples.   This tag does not grant access or share to FN4.",
			"tagAccessYN": "N",
			"tagRelatednessYN": "N"
		}, {
			"tagName": "test",
			"tagDescription": "This tag is for data pushed through as part of validation or other tests.",
			"tagAccessYN": "N",
			"tagRelatednessYN": "Y"
		}, {
			"tagName": "University_of_Oxford",
			"tagDescription": "Shares data to FN4 Shared Pool, does not grant external access",
			"tagAccessYN": "N",
			"tagRelatednessYN": "Y"

		}, {
			"tagName": "jpbarryustesting2022",
			"tagDescription": "jpbarryustesting2022",
			"tagAccessYN": "Y",
			"tagRelatednessYN": "Y"
		}]
	}]
}"""
    lib.parse_user_details(json.loads(json_payload))


def test_validate_fail_empty_country():
    """Empty country should fail"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-country.csv")
        )
    assert e.value.errors[0] == {
        "sample_name": "COVID_locost_2_barcode10",
        "error": "country cannot be empty",
    }


def test_validate_fail_empty_country_region():
    """Empty country should fail"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-country-region.csv")
        )
    assert e.value.errors[0] == {
        "sample_name": "COVID_locost_2_barcode10",
        "error": "country cannot be empty",
    }


def test_numeric_sample_name():
    """Pandera type coercion should mean that integer names are cast into strings"""
    df, schema_name = validation.validate(
        Path(data_dir) / Path("numeric-sample-name.csv")
    )


def test_run_number_generation():
    """
    Batches without any run numbers should return `sample: ""` for all samples
    Batches with one or more run numbers should provide run numbers for every sample
    """
    samples_runs_empty_single = {
        "sample1": "",
    }
    samples_runs_empty_multiple = {
        "sample1": "",
        "sample2": "",
    }
    samples_runs_full = {
        "sample1": "a",
        "sample2": "b",
    }
    samples_runs_mixed_single_run = {
        "sample1": "a",
        "sample2": "",
        "sample3": "a",
    }
    samples_runs_mixed_multiple_runs = {
        "sample1": "a",
        "sample2": "",
        "sample3": "b",
    }
    assert misc.number_runs(samples_runs_empty_single) == {"sample1": ""}
    assert misc.number_runs(samples_runs_empty_multiple) == {
        "sample1": "",
        "sample2": "",
    }
    assert misc.number_runs(samples_runs_full) == {"sample1": "1", "sample2": "2"}
    assert misc.number_runs(samples_runs_mixed_single_run) == {
        "sample1": "2",
        "sample2": "1",
        "sample3": "2",
    }
    assert misc.number_runs(samples_runs_mixed_multiple_runs) == {
        "sample1": "2",
        "sample2": "1",
        "sample3": "3",
    }


def test_populated_unpopulated_run_numbers():
    run_cmd = run("gpas upload populated-unpopulated-run-numbers.csv")


def test_populated_run_numbers():
    run_cmd = run("gpas upload populated-run-numbers.csv")


def test_unpopulated_run_numbers():
    run_cmd = run("gpas upload populated-run-numbers.csv")


def test_empty_sample_name_is_not_nan():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("empty-name.csv")
        )
    assert len(e.value.report["validation"]["errors"]) == 1
    assert "sample_name" not in e.value.report["validation"]["errors"][0]


def test_decontamination_extra_fields():
    lib.Batch(Path(data_dir) / Path("extra-fields.csv"))._decontaminate()


def test_decontamination_extra_fields_interleaved():
    lib.Batch(Path(data_dir) / Path("extra-fields-interleaved.csv"))._decontaminate()


def test_validate_json_messages_entire_output():
    """Will fail if …hypothetically… a stray print() is left in"""
    import datetime

    run_cmd = run("gpas validate broken/empty-date.csv --json-messages")
    assert run_cmd.stdout == (
        """{
    "validation": {
        "status": "failure",
        "errors": [
            {
                "sample_name": "COVID_locost_2_barcode10",
                "error": "collection_date cannot be empty"
            }
        ]
    }
}
"""
    )


def test_validate_colliding_mandatory_and_arbitrary_fields():
    df, schema = validation.validate(
        Path(data_dir) / Path("colliding-mandatory-arbitrary-fields.csv")
    )


def test_invalid_control():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("wrong-control.csv")
        )
    assert e.value.errors == [
        {
            "sample_name": "sample",
            "error": "heffalump in the control field is not valid; field must be either empty or contain one of the keywords negative, positive",
        },
    ]


def test_validate_fail_missing_platform():
    """Tests fix to KeyError in 0.8.0 due to calls to df["instrument_platform"] not being wrapped in try-except"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("missing-platform.csv")
        )


def test_validate_fail_missing_country():
    """Tests fix to KeyError in 0.8.0 due to calls to row["country"] not being wrapped in try-except"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("missing-country.csv")
        )


def test_validate_fail_missing_collection_date():
    """Purely prophylactic"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("missing-collection-date.csv")
        )


def test_validate_fail_missing_tags():
    """Purely prophylactic"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("missing-tags.csv")
        )


def test_validate_fail_missing_region():
    """Purely prophylactic"""
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("missing-region.csv")
        )



import asyncio
import gzip
import json
import subprocess
from pathlib import Path

import pytest
from gpas import lib, validation, misc
from gpas.misc import ENVIRONMENTS

data_dir = "tests/test-data"

auth = lib.parse_token(Path(data_dir) / Path("token.json"))
auth_result = lib.fetch_user_details(auth["access_token"], ENVIRONMENTS.dev)
_, _, permitted_tags, _ = lib.parse_user_details(auth_result)


def run(cmd, cwd="./"):  # Helper for CLI testing
    return subprocess.run(
        cmd, cwd=data_dir, shell=True, check=True, text=True, capture_output=True
    )


def test_status_guids_csv():
    run_cmd = run(
        f"gpas status --guids 6e024eb1-432c-4b1b-8f57-3911fe87555f,2ddbd7d4-9979-4960-8c17-e7b92f0bf413,8daadc7d-8d58-46a6-efb4-9ddefc1e4669 --format csv --environment dev token.json"
    )
    assert "6e024eb1-432c-4b1b-8f57-3911fe87555f,Unreleased" in run_cmd.stdout


def test_status_mapping_csv_json():
    run_cmd = run(
        f"gpas status --mapping-csv example-mapping-csv.csv --format json --environment dev token.json"
    )
    assert {
        "sample": "6e024eb1-432c-4b1b-8f57-3911fe87555f",
        "status": "Unreleased",
    } in json.loads(run_cmd.stdout)


def test_status_mapping_csv_csv():
    run_cmd = run(
        f"gpas status --mapping-csv example-mapping-csv.csv --format csv --environment dev token.json"
    )
    assert "6e024eb1-432c-4b1b-8f57-3911fe87555f,Unreleased" in run_cmd.stdout


def test_status_mapping_csv_table():
    run_cmd = run(
        f"gpas status --mapping-csv example-mapping-csv.csv --format table --environment dev token.json"
    )
    assert "6e024eb1-432c-4b1b-8f57-3911fe87555f Unreleased" in run_cmd.stdout


def test_status_mapping_csv_rename():
    run_cmd = run(
        f"gpas status --mapping-csv example-mapping-csv.csv --environment dev token.json --format csv --rename"
    )
    assert "test4_uploaded,Uploaded" in run_cmd.stdout


def test_gpas_uploader_download_mapping_csv():
    run_cmd = run(
        f"gpas download --mapping-csv example-mapping-csv.csv --environment dev token.json"
    )
    assert Path(f"{data_dir}/6e024eb1-432c-4b1b-8f57-3911fe87555f.fasta.gz").is_file()
    run("rm -f *.fasta.gz")


def test_download_mapping_csv():
    run_cmd = run(
        f"gpas download --mapping-csv example-mapping-csv.csv --environment dev token.json"
    )
    assert Path(f"{data_dir}/6e024eb1-432c-4b1b-8f57-3911fe87555f.fasta.gz").is_file()
    run("rm -f *.fasta.gz")


def test_download_mapping_csv_rename():
    run_cmd = run(
        f"gpas download --rename --mapping-csv example-mapping-csv.csv --file-types vcf --environment dev token.json"
    )
    assert Path(f"{data_dir}/test1.vcf").is_file()
    run("rm -f test*.vcf")


def test_download_mapping_csv_rename_out_dir():
    run
    run_cmd = run(
        f"mkdir -p outdir && gpas download --rename --mapping-csv example-mapping-csv.csv --file-types fasta,vcf,json --environment dev token.json --out-dir outdir"
    )
    assert Path(f"{data_dir}/outdir/test1.fasta.gz").is_file()
    run("rm -rf outdir")


def test_download_guid_rename_without_mapping():
    run_cmd = run(
        f"gpas download --guids 6e024eb1-432c-4b1b-8f57-3911fe87555f --file-types vcf --rename --environment dev token.json"
    )
    assert "Cannot rename outputs without mapping CSV" in run_cmd.stderr
    assert Path(f"{data_dir}/6e024eb1-432c-4b1b-8f57-3911fe87555f.vcf").is_file()
    run("rm -f 6e024eb1-432c-4b1b-8f57-3911fe87555f.vcf")


def test_download_guid_api():
    auth = lib.parse_token(Path(data_dir) / Path("token.json"))
    asyncio.run(
        lib.download_async(
            guids=["6e024eb1-432c-4b1b-8f57-3911fe87555f"],
            file_types=["vcf"],
            access_token=auth["access_token"],
            out_dir=data_dir,
            environment=ENVIRONMENTS.dev,
        )
    )
    assert (Path(data_dir) / Path("6e024eb1-432c-4b1b-8f57-3911fe87555f.vcf")).is_file()
    run("rm -f 6e024eb1-432c-4b1b-8f57-3911fe87555f.vcf")


def test_status_mapping_api():
    access_token = lib.parse_token(Path(data_dir) / Path("token.json"))["access_token"]
    guids_names = lib.parse_mapping_csv(
        Path(data_dir) / Path("example-mapping-csv.csv")
    )
    records = asyncio.run(
        lib.fetch_status_async(
            access_token=access_token,
            guids=guids_names.keys(),
            environment=ENVIRONMENTS.dev,
        )
    )
    assert records  # Smoke test
    passed = False
    for r in records:
        if (
            r["sample"] == "8daadc7d-8d58-46a6-efb4-9ddefc1e4669"
            and r["status"] == "Uploaded"
        ):
            passed = True
    if not passed:
        raise RuntimeError("Expected dict not found")


def test_status_sync_mapping():
    access_token = lib.parse_token(Path(data_dir) / Path("token.json"))["access_token"]
    guids_names = lib.parse_mapping_csv(
        Path(data_dir) / Path("example-mapping-csv.csv")
    )
    records = lib.fetch_status(
        access_token=access_token,
        guids=guids_names.keys(),
        environment=ENVIRONMENTS.dev,
    )
    print(records)
    assert records  # Smoke test
    passed = False
    for r in records:
        if (
            r["sample"] == "8daadc7d-8d58-46a6-efb4-9ddefc1e4669"
            and r["status"] == "Uploaded"
        ):
            passed = True
    if not passed:
        raise RuntimeError("Expected dict not found")


# pytest -s --online tests/test_gpas_online.py::test_status_sync_mapping_rename
def test_status_sync_mapping_rename():
    access_token = lib.parse_token(Path(data_dir) / Path("token.json"))["access_token"]
    guids_names = lib.parse_mapping_csv(
        Path(data_dir) / Path("example-mapping-csv.csv")
    )
    records = lib.fetch_status(
        access_token=access_token,
        guids=guids_names,
        environment=ENVIRONMENTS.dev,
    )
    print(records)
    assert records  # Smoke test
    passed = False
    for r in records:
        if r["sample"] == "test4_uploaded" and r["status"] == "Uploaded":
            passed = True
    if not passed:
        raise RuntimeError("Expected dict not found")


def test_status_mapping_rename_api():
    access_token = lib.parse_token(Path(data_dir) / Path("token.json"))["access_token"]
    guids_names = lib.parse_mapping_csv(
        Path(data_dir) / Path("example-mapping-csv.csv")
    )
    records = asyncio.run(
        lib.fetch_status_async(
            access_token=access_token,
            guids=guids_names,
            environment=ENVIRONMENTS.dev,
        )
    )
    assert records  # Smoke test
    passed = False
    for r in records:
        if r["sample"] == "test4_uploaded" and r["status"] == "Uploaded":
            passed = True
    if not passed:
        raise RuntimeError("Expected dict not found")


def test_gpas_uploader_download_mapping_rename_fasta():
    run_cmd = run(
        f"gpas download --mapping-csv example-mapping-csv.csv --rename --environment dev token.json"
    )
    with gzip.open(Path(f"{data_dir}/test1.fasta.gz"), "rt") as fh:
        assert "cdbc4af8-a75c-42ce-8fe2-8dba2ab5e839|test1" in fh.read()
    run("rm -f *.fasta.gz")


def test_check_auth_success():
    auth = lib.parse_token(Path(data_dir) / Path("token.json"))
    auth_result = lib.fetch_user_details(auth["access_token"], ENVIRONMENTS.dev)
    lib.parse_user_details(auth_result)


def test_gpas_validate():
    """Check that stderr says successful but that stdout is empty"""
    run_cmd = run(
        f"gpas validate --token token.json --environment dev large-nanopore-fastq.csv"
    )
    assert "success" in run_cmd.stderr
    assert run_cmd.stdout == ""


def test_gpas_validate_message():
    """Check that"""
    run_cmd = run(
        f"gpas validate --token token.json --environment dev large-nanopore-fastq.csv --json-messages"
    )
    assert '"status": "success"' in run_cmd.stdout


def test_validate_fail_wrong_tags():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            upload_csv=Path(data_dir) / Path("broken") / Path("wrong-tags.csv"),
            permitted_tags=permitted_tags,
        )
    assert e.value.errors == [
        {"error": "tag(s) {'heffalump'} are unauthorised for this user"}
    ]


def test_validate_fail_wrong_tags_cli():
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd = run(
            f"gpas validate --environment dev --token token.json broken/wrong-tags.csv"
        )


def test_validate_fail_no_tags():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("no-tags.csv"),
            permitted_tags=permitted_tags,
        )
    assert e.value.errors == [
        {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be empty"}
    ]


def test_validate_fail_no_tags_colon():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("no-tags-colon.csv"),
            permitted_tags=permitted_tags,
        )
    assert e.value.errors == [
        {"sample_name": "cDNA-VOC-1-v4-1", "error": "tags cannot be empty"}
    ]


def test_validate_fail_valid_and_invalid_tags():
    with pytest.raises(validation.ValidationError) as e:
        _, message = validation.validate(
            Path(data_dir) / Path("broken") / Path("valid-and-invalid-tags.csv"),
            permitted_tags=permitted_tags,
        )
    assert (
        "tag(s) {'heffalump'} are unauthorised for this user"
        in e.value.errors[0]["error"]
    )


def test_validate_valid_and_empty_tags():
    """Empty tags are trimmed so should work fine"""
    validation.validate(
        Path(data_dir) / Path("valid-and-empty-tags.csv"),
        permitted_tags=permitted_tags,
    )


def test_fail_auth_wrong_token():
    with pytest.raises(misc.AuthenticationError):
        auth = lib.parse_token(Path(data_dir) / Path("token.json"))
        result = lib.fetch_user_details(auth["access_token"], ENVIRONMENTS.prod)
        _, _, permitted_tags, _ = lib.parse_user_details(result)


def test_auth_broken_token():
    with pytest.raises(misc.AuthenticationError):
        broken_auth = lib.parse_token(
            Path(data_dir) / Path("broken") / Path("broken-token.json")
        )
        _, _, permitted_tags, _ = lib.fetch_user_details(
            broken_auth["access_token"], ENVIRONMENTS.dev
        )


def test_ont_bam_dry():
    run_cmd = run(
        f"gpas upload --environment dev --token token.json large-nanopore-bam.csv --dry-run"
    )
    assert "INFO: Finished converting 1 sample(s)" in run_cmd.stderr
    assert "INFO: Finished decontaminating 1 sample(s)" in run_cmd.stderr



import datetime
import json
import logging
import os
import re
from pathlib import Path

import pandas as pd
import pandera as pa
import pandera.extensions as extensions
from pandera.typing import Index, Series

import gpas


def parse_countries_subdivisions():
    """Parse static JSON generated by Pycountry to avoid dependency"""
    path = gpas.data_dir / Path("countries") / Path("countries_subdivisions.json")
    if not path.exists():
        path = (
            Path(__file__).parent.parent
            / Path("data")
            / Path("countries")
            / Path("countries_subdivisions.json")
        )
    with open(path, encoding="utf-8") as fh:
        return json.load(fh)


CONTROLS = {"positive", "negative"}
HOSTS = {"human"}
INSTRUMENTS = {"Illumina", "Nanopore"}
ORGANISMS = {"SARS-CoV-2"}
PRIMER_SCHEMES = {"auto"}
COUNTRIES_SUBDIVISIONS = parse_countries_subdivisions()
COUNTRIES_ALPHA_3 = COUNTRIES_SUBDIVISIONS.keys()
REGIONS = {i for l in COUNTRIES_SUBDIVISIONS.values() for i in l}


class ValidationError(Exception):
    def __init__(self, errors):
        self.errors = errors
        self.errors_df = pd.DataFrame(errors, columns=["sample_name", "error"]).fillna(
            value=""
        )
        self.report = {
            "validation": {
                "status": "failure",
                "errors": errors,
            }
        }
        super().__init__(self._message())

    def _message(self):
        message = f"Failed to validate upload CSV ({len(self.errors)} errors):\n"
        message += self.errors_df.to_string(index=False, justify="left")
        return message


class set_directory(object):
    """
    Context manager for temporarily changing the current working directory
    """

    def __init__(self, path: Path):
        self.path = path
        self.origin = Path().absolute()

    def __enter__(self):
        os.chdir(self.path)

    def __exit__(self, *exc):
        os.chdir(self.origin)


class BaseSchema(pa.SchemaModel):
    """
    Validate generic GPAS upload CSVs
    """

    # validate that batch is alphanumeric only
    batch: Series[str] = pa.Field(str_matches=r"^[A-Za-z0-9._-]+$", nullable=True)

    # validate run_number is alphanumeric but can also be null
    run_number: Series[str] = pa.Field(
        str_matches=r"^[A-Za-z0-9._-]+$",
        nullable=True,
    )

    # validate sample name is alphanumeric and insist it is unique
    sample_name: Index[str] = pa.Field(
        str_matches=r"^[A-Za-z0-9._-]+$",
        unique=True,
        nullable=False,
    )

    # insist that control is one of positive, negative or null
    control: Series[str] = pa.Field(
        nullable=True,
        isin=CONTROLS,
    )

    # validate that the collection is in the ISO format, is no earlier than 01-Jan-2019 and no later than today
    collection_date: Series[str] = pa.Field(
        nullable=False,
    )

    # insist that the tags is alphanumeric, including : as it is the delimiter
    tags: Series[str] = pa.Field(
        nullable=False,
        str_matches=r"^[A-Za-z0-9:_-]+$",
    )

    # insist that the country is one of the entries in the specified lookup table
    country: Series[str] = pa.Field(isin=COUNTRIES_ALPHA_3, nullable=False)

    region: Series[str] = pa.Field(nullable=True, isin=REGIONS)

    district: Series[str] = pa.Field(
        str_matches=r"^[\sA-Za-z0-9:_-]+$",
        nullable=True,
    )

    # at present specimen_organism can only be SARS-CoV-2
    specimen_organism: Series[str] = pa.Field(isin=ORGANISMS, nullable=False)

    # at present host can only be human
    host: Series[str] = pa.Field(isin=HOSTS, nullable=False)

    # insist that instrument_platform can only be Illumina or Nanopore
    instrument_platform: Series[str] = pa.Field(isin=INSTRUMENTS, nullable=False)

    # at present primer_schema can only be auto
    primer_scheme: Series[str] = pa.Field(isin=PRIMER_SCHEMES, nullable=False)

    @pa.check(collection_date, element_wise=True)
    def check_collection_date(cls, value: str) -> bool:
        """Check the date is parseable and within allowed range"""
        valid = True
        try:
            date = pd.to_datetime(value, format="%Y-%m-%d")
            is_parseable = date.strftime("%Y-%m-%d") == value
        except Exception:
            is_parseable = False
        try:
            allowed_range_start = datetime.datetime.strptime("2019-01-01", "%Y-%m-%d")
            allowed_range_end = datetime.datetime.today()
            is_within_range = allowed_range_start < date < allowed_range_end
        except Exception:
            is_within_range = False
        return valid if is_parseable and is_within_range else False

    @pa.check(instrument_platform)
    def check_unique_instrument_platform(cls, field):
        """
        Check that only one instrument_platform is specified in the upload CSV
        """
        return len(field.unique()) == 1

    @pa.check(tags, element_wise=True)
    def tags_are_unique(cls, value: str) -> bool:
        valid = True
        if value and not pd.isna(value):
            value = value.strip(":")
            if len(set(value.split(":"))) != len(list(value.split(":"))):
                valid = False
        return valid

    @pa.check(tags, element_wise=True)
    def tags_are_present(cls, value: str) -> bool:
        """Catch colon-only case"""
        return bool(str(value).strip(":"))

    @pa.dataframe_check(ignore_na=False)
    def region_is_valid(cls, df: pd.DataFrame) -> Series[bool]:
        """
        Validate the region field using ISO 3166-2
        """

        def validate_region(row):
            try:
                region = row["region"] if "region" in row else ""
                if (
                    region
                    and not pd.isna(region)
                    and region not in COUNTRIES_SUBDIVISIONS.get(row["country"], {})
                ):
                    valid = False
                else:
                    valid = True
            except KeyError:
                valid = False
            return valid

        return df.apply(validate_region, axis=1)

    class Config:
        coerce = True


class FastqSchema(BaseSchema):
    """
    Validate GPAS upload CSVs specifying unpaired reads (e.g. Nanopore).
    """

    # validate that the fastq file is alphanumeric and unique
    fastq: Series[str] = pa.Field(
        unique=True,
        str_matches=r"^[A-Za-z0-9 /._-]+$",
        str_endswith=".fastq.gz",
        nullable=False,
    )

    @pa.check(fastq, element_wise=True)
    def check_path_fastq(cls, path: str) -> bool:
        if path and pd.notna(path):
            return Path(path).exists()


class PairedFastqSchema(BaseSchema):
    """
    Validate GPAS upload CSVs specifying paired reads (e.g Illumina).
    """

    fastq1: Series[str] = pa.Field(
        unique=True,  # Joint uniqueness specified in Config
        str_matches=r"^[A-Za-z0-9 /._-]+$",
        str_endswith=".fastq.gz",
        nullable=False,
    )

    fastq2: Series[str] = pa.Field(
        unique=True,  # Joint uniqueness specified in Config
        str_matches=r"^[A-Za-z0-9 /._-]+$",
        str_endswith=".fastq.gz",
        nullable=False,
    )

    @pa.check(fastq1, element_wise=True)
    def check_path_fastq1(cls, path: str) -> bool:
        """Check that fastq1 exists"""
        if path and pd.notna(path):
            return Path(path).exists()

    @pa.check(fastq2, element_wise=True)
    def check_path_fastq2(cls, path: str) -> bool:
        """Check that fastq2 exists"""
        if path and pd.notna(path):
            return Path(path).exists()

    @pa.dataframe_check(ignore_na=False)
    def fastq1_does_not_equal_fastq2(cls, df: pd.DataFrame) -> Series[bool]:
        """Check that fastq1 and fastq2 are not the same file"""
        return df["fastq1"] != df["fastq2"]

    class Config:
        unique = ["fastq1", "fastq2"]


class BamSchema(BaseSchema):
    """
    Validate GPAS upload CSVs specifying BAM files.
    """

    bam: Series[str] = pa.Field(
        unique=True,
        str_matches=r"^[A-Za-z0-9 /._-]+$",
        str_endswith=".bam",
        nullable=False,
    )

    @pa.check(bam, element_wise=True)  # Check that bam exists
    def check_path_bam(cls, path: str) -> bool:
        if path and pd.notna(path):
            return Path(path).exists()


class PairedBamSchema(BamSchema):
    pass


def get_valid_samples(df: pd.DataFrame, schema_name: str) -> list[dict]:
    samples = []
    for row in df.reset_index().itertuples():
        if schema_name == "FastqSchema":
            samples.append({"sample_name": row.sample_name, "files": [row.fastq]})
        elif schema_name == "PairedFastqSchema":
            samples.append(
                {"sample_name": row.sample_name, "files": [row.fastq1, row.fastq2]}
            )
        elif schema_name in {"BamSchema", "PairedBamSchema"}:
            samples.append({"sample_name": row.sample_name, "files": [row.bam]})
        else:
            raise ValidationError([{"error": "Unexpected schema"}])
    return samples


def remove_ints(ld: list[dict]):
    """Eliminate autogenerated integer sample names for empty samples"""
    return [{k: v for k, v in d.items() if type(v) != int} for d in ld]


def remove_nones_duplicates_empties_ints_from_ld(ld: list[dict]) -> list[dict]:
    """Remove None-valued keys, duplicated, and empty dicts from a list of dicts"""
    ld = (
        pd.DataFrame(ld, columns=["sample_name", "error"])
        .fillna(value="")
        .sort_values(["sample_name", "error"])
        .drop_duplicates()
        .to_dict("records")
    )
    return list(filter(None, [{k: v for k, v in d.items() if v} for d in ld]))


def parse_validation_errors(errors):
    """Parse errors arising during Pandera SchemaModel validation

    Parameters
    ----------
    err: pa.errors.SchemaErrors

    Returns
    -------
    pandas.DataFrame(columns=['sample_name', 'error_message'])
    """
    failure_cases = errors.failure_cases.rename(columns={"index": "sample_name"})
    failure_cases["error"] = failure_cases.apply(parse_validation_error, axis=1)
    failures = failure_cases[["sample_name", "error"]].to_dict("records")
    return remove_nones_duplicates_empties_ints_from_ld(remove_ints(failures))


def parse_validation_error(row):
    """
    Generate palatable errors from pandera output
    """
    # print(str(row), "\n")
    if row.check == "column_in_schema":
        return "unexpected column " + row.failure_case
    if row.check == "column_in_dataframe":
        return "missing column " + row.failure_case
    elif row.check == "region_is_valid":
        return "invalid region (ISO 3166-2 subdivision) for specified country"
    elif row.check == "instrument_is_valid" or (
        row.column == "instrument_platform" and "isin" in row.check
    ):
        return (
            f"instrument_platform can only contain one of {list(sorted(INSTRUMENTS))}"
        )
    elif row.check == "not_nullable":
        if row.schema_context == "Index":
            return "sample_name cannot be empty"
        else:
            return row.column + " cannot be empty"
    elif row.check == "field_uniqueness":
        if row.schema_context == "Index":
            return "sample_name must be unique"
        else:
            return row.column + " must be unique"
    elif row.check == "multiple_fields_uniqueness":
        return "fastq1 and fastq2 must be jointly unique"
    elif row.check == "check_collection_date":
        return (
            row.column
            + " must be in format YYYY-MM-DD between 2019-01-01 and "
            + datetime.date.today().isoformat()
        )
    elif row.check == "fastq1_does_not_equal_fastq2":
        return "fastq1 and fastq2 cannot be the same"
    elif "str_matches" in row.check:
        allowed_chars = row.check.split("[")[1].split("]")[0]
        if row.schema_context == "Column":
            if row.column == "district" and row.failure_case == False:
                return None  # empty-name.csv causes problems for district
            else:
                return (
                    row.column + " can only contain characters (" + allowed_chars + ")"
                )
        elif row.schema_context == "Index":
            return "sample_name can only contain characters (" + allowed_chars + ")"
    elif row.column == "country" and row.check[:4] == "isin":
        return row.failure_case + " is not a valid ISO 3166-1 alpha-3 country code"
    elif row.column == "region" and row.check[:4] == "isin":
        return row.failure_case + " is not a valid ISO 3166-2 subdivision name"
    elif row.column == "control" and row.check[:4] == "isin":
        if row.failure_case == False:  # empty-name.csv causes problems for controls
            return None
        else:
            return (
                row.failure_case
                + f" in the control field is not valid; field must be either empty or contain one of the keywords {', '.join(sorted(CONTROLS))}"
            )
    elif row.column == "host" and row.check[:4] == "isin":
        return row.column + " can only contain the keyword human"
    elif row.column == "specimen_organism" and row.check[:4] == "isin":
        return row.column + " can only contain the keyword SARS-CoV-2"
    elif row.column == "primer_scheme" and row.check[:4] == "isin":
        return row.column + " can only contain the keyword auto"
    elif row.column == "instrument_platform":
        return row.column + " must be the same for all samples in a submission"
    elif row.column is None:
        return "problem"
    elif row.check == "tags_are_unique":
        return row.column + " cannot be repeated"
    elif row.check == "tags_are_present":
        return row.column + " cannot be empty"
    elif row.check.startswith("check_path"):
        return row.column + " file does not exist"
    elif row.check.startswith("str_endswith"):
        return row.column + " must end with .fastq.gz or .bam as appropriate"
    else:
        return "problem in " + row.column + " field"


def select_schema(df: pd.DataFrame) -> pa.SchemaModel:
    """Choose appropriate validation schema and the presence of required columns"""
    columns = set(df.columns)
    try:
        if "bam" in columns and not {"fastq", "fastq1", "fastq2"} & columns:
            if "Illumina" in df["instrument_platform"].tolist():
                schema = PairedBamSchema
            else:
                schema = BamSchema
        elif (
            "fastq" in columns
            and not {"fastq1", "fastq2", "bam"} & columns
            and "Nanopore" in df["instrument_platform"].tolist()
        ):
            schema = FastqSchema
        elif (
            {"fastq1", "fastq2"} < columns
            and not {"fastq", "bam"} & columns
            and "Illumina" in df["instrument_platform"].tolist()
        ):
            schema = PairedFastqSchema
        else:
            raise RuntimeError()
    except:
        raise ValidationError(
            [
                {
                    "error": "could not infer upload CSV schema."
                    " For Nanopore samples, column 'instrument_platform' must be"
                    " 'Nanopore', and either column 'fastq' or column 'bam' must be"
                    " valid paths."
                    " For Illumina samples, column 'instrument_platform' must be"
                    " 'Illumina' and either columns 'bam' or 'fastq1' and 'fastq2' must"
                    " be valid paths."
                }
            ]
        )
    return schema


def resolve_paths(df: pd.DataFrame) -> pd.DataFrame:
    """
    Resolve relative paths to files inside dataframe
    """
    resolve = lambda x: str(Path(x).resolve())
    if "fastq" in df.columns:
        df["fastq"] = df["fastq"].apply(resolve)
    if "fastq1" in df.columns:
        df["fastq1"] = df["fastq1"].apply(resolve)
    if "fastq2" in df.columns:
        df["fastq2"] = df["fastq2"].apply(resolve)
    if "bam" in df.columns:
        df["bam"] = df["bam"].apply(resolve)
    return df


def validate_tags(df, permitted_tags):
    """
    Validate tags in upload csv
    """
    invalid_tags = set()
    for value in df["tags"].tolist():  # Get tags for each record
        if value and not pd.isna(value):
            tags = set(filter(None, value.strip(" :").split(":")))  # Get non-empty tags
            for tag in tags:
                if tag not in permitted_tags:
                    invalid_tags.add(tag)

    if invalid_tags:
        raise ValidationError(
            [{"error": f"tag(s) {invalid_tags} are unauthorised for this user"}]
        )


def validate_upload_csv_path(upload_csv) -> bool:
    return bool(re.match(r"^[A-Za-z0-9\\\s:./_-]+$", str(upload_csv)))


def validate(
    upload_csv: Path, permitted_tags: list[str] = []
) -> tuple[pd.DataFrame, pa.SchemaModel]:
    """
    Validate a dataframe. Returns success message or throws ValidationError
    """
    if not validate_upload_csv_path(upload_csv):
        raise ValidationError(
            [{"error": "upload csv path contains illegal characters"}]
        )
    try:
        df = pd.read_csv(
            upload_csv,
            encoding="utf-8",
            index_col="sample_name",
            dtype=str,
        )
    except Exception as e:
        raise ValidationError(
            [{"error": f"failed to parse upload CSV ({str(e)})"}]
        ) from None

    schema = select_schema(df)
    if permitted_tags:  # Only validate if we have tags
        validate_tags(df, permitted_tags)
    try:
        with set_directory(Path(upload_csv).parent):
            df = resolve_paths(schema.validate(df, lazy=True))
    except pa.errors.SchemaErrors as e:  # Validation errorS, because lazy=True
        raise ValidationError(parse_validation_errors(e)) from None
    logging.info("Validation successful")
    return df, schema


def build_validation_message(df: pd.DataFrame, schema_name: str) -> dict:
    records = get_valid_samples(df, schema_name)
    message = {
        "validation": {
            "status": "success",
            "samples": records,
        }
    }
    return message
